#pragma once

#include <amdis/LinearAlgebra.hpp>
#include <amdis/OperatorList.hpp>
#include <amdis/common/FlatVector.hpp>
#include <amdis/common/TypeTraits.hpp>
#include <amdis/typetree/TreePath.hpp>

namespace AMDiS
{
  /// \brief The basic container that stores a base vector and a corresponding basis
  /**
   * A vector storing all the assembled Operators indexed with DOF indices. The vector
   * data is associated to a global basis.
   *
   * \tparam GB  Basis of the vector
   * \tparam T   Coefficient type to store in the vector
   **/
  template <class GB, class T = double>
  class LinearForm
      : public VectorBase<GB, VectorBackend<BackendTraits<GB,T>>>
  {
    using Super = VectorBase<GB, VectorBackend<BackendTraits<GB,T>>>;

  public:
    using Backend = VectorBackend<BackendTraits<GB,T>>;
    using Comm = typename Backend::Traits::Comm;

    /// The type of the functionspace basis associated to this linearform
    using GlobalBasis = GB;

    /// A traits class collecting several parameters of the linear algebra backend
    using Traits = typename Backend::Traits;

    /// The type of the elements of the DOFVector
    using CoefficientType = typename Traits::CoefficientType;

    /// The type of the vector filled on an element with local contributions
    using ElementVector = FlatVector<CoefficientType>;

  public:
    /// Constructor. Stores the shared_ptr of the basis and creates a new DataTransfer.
    template <class GB_, class Comm_>
    LinearForm(GB_&& basis, Comm_&& comm)
      : Super(FWD(basis), FWD(comm))
    {
      operators_.init(*Super::basis());
    }

    /// Prepare the LinearForm for insertion of values, finish the insertion with
    /// \ref finish().
    void init(bool clear)
    {
      Super::init(clear);

      auto const localSize = this->basis()->localView().maxSize();
      elementVector_.resize(localSize);
    }

    /// Associate a local operator with this LinearForm
    template <class ContextTag, class Expr, class TreePath = RootTreePath>
    void addOperator(ContextTag contextTag, Expr const& expr, TreePath path = {});

    /// Assemble the vector operators on the bound element.
    void assemble(typename GB::LocalView const& localView);

    /// Assemble all vector operators added by \ref addOperator().
    void assemble();

  private:
    /// Dense vector to store coefficients during \ref assemble()
    ElementVector elementVector_;

    /// List of operators associated to nodes, filled in \ref addOperator().
    VectorOperators<GlobalBasis,ElementVector> operators_;
  };

#if DUNE_HAVE_CXX_CLASS_TEMPLATE_ARGUMENT_DEDUCTION
  template <class GB, class C>
  LinearForm(GB&& basis, C const& comm)
    -> LinearForm<Underlying_t<GB>, VectorBackend<BackendTraits<Underlying_t<GB>>> >;
#endif

} // end namespace AMDiS

#include <amdis/LinearForm.inc.hpp>
